/*****************************************************************************
* Product: QP-nano public interface
* Last Updated for Version: 1.1.00
* Date of the Last Update:  May 13, 2006
*
*                    Q u a n t u m     L e a P s
*                    ---------------------------
*                    innovating embedded systems
*
* Copyright (c) 2002-2006 Quantum Leaps, LLC. All rights reserved.
*
* Internet: www.quantum-leaps.com     Licensing: sales@quantum-leaps.com
*
* This Software is protected by the United States copyright laws and
* international treaties. Distribution of products containing this Software
* or based upon this Software (Derivative Works) requires a valid Quantum
* Leaps Distribution License. Any other distribution, in source or binary
* format is illegal.
*****************************************************************************/
#ifndef qpn_h
#define qpn_h

/** \ingroup qepn qfn
* \file qpn.h
* \brief Public QP-nano interface.
*
* This header file must be included in all modules that use QP-nano.
* Typically, this header file is included indirectly through the
* header file qpn_port.h.
*/

#include <stdint.h>                        /* C99 exact-width integer types */

/****************************************************************************/
/** helper macro to calculate static dimension of a 1-dim array \a array_ */
#define Q_DIM(array_) (sizeof(array_) / sizeof(array_[0]))

/****************************************************************************/
/** \brief get the current QP version number string
*
* \return version of the QP as a constant 6-character string of the form
* x.y.zz, where x is a 1-digit major version number, y is a 1-digit minor
* version number, and zz is a 2-digit release number.
*/
char const Q_ROM *QP_getVersion(void);

/** \brief Scalar type describing the signal of an event.
*/
typedef uint8_t QSignal;

/****************************************************************************/
#ifndef Q_PARAM_SIZE
    /** \brief macro to define the size of event parameter.
    * Valid values 0, 1, 2, or 4; default 0
    */
    #define Q_PARAM_SIZE 0
#endif
#if (Q_PARAM_SIZE == 0)
#elif (Q_PARAM_SIZE == 1)

    /** \brief type of the event parameter.
    *
    * This typedef is configurable via the preprocessor switch #Q_PARAM_SIZE.
    * The other possible values of this type are as follows: \n
    * none when (Q_PARAM_SIZE == 0); \n
    * uint8_t when (Q_PARAM_SIZE == 1); \n
    * uint16_t when (Q_PARAM_SIZE == 2); and \n
    * uint32_t when (Q_PARAM_SIZE == 4).
    */
    typedef uint8_t QParam;
#elif (Q_PARAM_SIZE == 2)
    typedef uint16_t QParam;
#elif (Q_PARAM_SIZE == 4)
    typedef uint32_t QParam;
#else
    #error "Q_PARAM_SIZE defined incorrectly, expected 0, 1, 2, or 4"
#endif

/** \brief a name for QEventTag struct.
*
* QEvent represents events, optionally with a single scalar parameter.
* \sa Q_PARAM_SIZE
* \sa ::QParam
*/
typedef struct QEventTag QEvent;
/** \brief Event structure
* \sa ::QEvent
*/
struct QEventTag {
    QSignal sig;                                   /**< signal of the event */

#if (Q_PARAM_SIZE != 0)
    QParam par;                          /**< scalar parameter of the event */
#endif
};

/****************************************************************************/
/** \brief QP reserved signals */
enum QReservedSignals {
    Q_ENTRY_SIG = 1,                   /**< signal for coding entry actions */
    Q_EXIT_SIG,                         /**< signal for coding exit actions */
    Q_INIT_SIG,           /**< signal for coding nested initial transitions */
    Q_TIMEOUT_SIG,                          /**< signal used by time events */
    Q_USER_SIG      /**< first signal that can be used in user applications */
};

/** signal for coding the top-most initial transition */
#define Q_TOP_INIT_SIG    Q_INIT_SIG


/****************************************************************************/
/** \brief a name for QFsmTag struct
*
* QFsm represents a traditional non-hierarchical Finite State Machine (FSM)
* without state hierarchy, but with entry/exit actions.
*
* \note QFsm is not intended to be instantiated directly, but rather serves
* as the base structure for derivation of state machines in the application
* code.
*
* The following example illustrates how to derive a state machine structure
* from QFsm. Please note that the QFsm member super_ is defined as the FIRST
* member of the derived struct.
* \include qepn_qfsm.c
*
* \sa \ref derivation
*/
typedef struct QFsmTag QFsm;

                         /** \brief the signature of state handler function */
typedef void (*QState)(QFsm *me);

/** \brief Finite State Machine
* \sa ::QFsm
*/
struct QFsmTag {
    QEvent evt_;      /**< currently processed event in the FSM (protected) */
    QState state__;          /**< current active state of the FSM (private) */
};

/** \brief macro to access the signal of the current event of a state machine
*
* \sa ::QFsm ::QHsm
*/
#define Q_SIG(me_)  (((QFsm *)(me_))->evt_.sig)

#if (Q_PARAM_SIZE != 0)
/** \brief macro to access the parameter of the current event of
* a state machine
*
* \sa ::QFsm ::QHsm Q_PARAM_SIZE
*/
#define Q_PAR(me_)  (((QFsm *)(me_))->evt_.par)
#endif

#ifndef Q_NFSM

/** \brief Performs the second step of FSM initialization by triggering the
* top-most initial transition.
*
* \param me_ is the pointer the state machine structure derived from ::QFsm.
* \param initial_ is the pointer to the initial state of the state machine.
* \note Must be called only ONCE before dispatching any events via
* QFsm_dispatch().
*/
#define QFsm_init(me_, initial_) \
    (((QFsm *)me_)->state__ = (QState)(initial_), \
     (me_)->evt_.sig = (QSignal)Q_ENTRY_SIG, \
     QFsm_dispatch(me_))

/** \brief Dispatches an event to a FSM
*
* Processes one event at a time in Run-to-Completion fashion. The argument
* \a me is the pointer the state machine structure derived from ::QFsm.
*
* \note Must be called after QFsm_init().
*/
void QFsm_dispatch(QFsm *me);

#endif                                                            /* Q_NFSM */

/* protected methods */

/** \brief Returns current active state of a FSM.
*
* \note this is a protected function to be used only inside state handler
* functions.
*/
#define QFsm_getState_(me_) ((QState const)(me_)->super_.state__)

/** \brief Designates a target for an initial transition.
*
* Q_INIT() should be used exclusively in the nested initial transitions
* triggered by ::Q_INIT_SIG, but never in regular transitions, including
* the top-most initial transition triggered by ::Q_TOP_INIT_SIG.
*
* The following example demonstrates the use of Q_INIT():
* \include qepn_qinit.c
*/
#define Q_INIT(target_) (((QFsm *)me)->state__ = (QState)(target_))

/** \brief Designates a target for an regular transition.
*
* Q_TRAN() can be used both in the FSMs and HSMs:
*
* \include qepn_qtran.c
*/
#define Q_TRAN(target_) \
    (((QFsm *)me)->state__ = (QState)(target_), \
     ((QFsm *)me)->evt_.sig = (QSignal)0)

/****************************************************************************/
#ifndef Q_NHSM

/** \brief a name for QHsmTag struct
*
* QHsm represents a Hierarchical Finite State Machine (HSM). QHsm
* extends the capabilities of a basic FSM with state hierarchy.
*
* \note QHsm is not intended to be instantiated directly, but rather serves
* as the base structure for derivation of state machines in the application
* code.
*
* The following example illustrates how to derive a state machine structure
* from QHsm. Please note that the QHsm member super_ is defined as the FIRST
* member of the derived struct.
* \include qepn_qhsm.c
*
* \sa \ref derivation
*/
typedef struct QHsmTag QHsm;

                 /** \brief the signature of state handler function for HSM */
typedef QState (*QHsmState)(QHsm *me);

  /** \brief a name for the return type from the HSM state handler function */
typedef QState QSTATE;

/** \brief Hierarchical State Machine
* \sa ::QHsm
*/
struct QHsmTag {
    QEvent evt_;      /**< currently processed event in the HSM (protected) */
    QHsmState state__;       /**< current active state of the HSM (private) */
};

/* public methods */
/** \brief Performs the second step of HSM initialization by triggering the
* top-most initial transition.
*
* \param me_ pointer the state machine structure derived from ::QHsm.
* \param initial_ is the pointer to the initial state of the state machine.
* \note Must be called only ONCE before dispatching any events via
* QHsm_dispatch().
*/
#define QHsm_init(me_, initial_) \
    (((QHsm *)me_)->state__ = (QHsmState)(initial_), \
     (me_)->evt_.sig = (QSignal)Q_TOP_INIT_SIG, \
     QHsm_dispatch(me_))

/** \brief Dispatches an event to a HSM
*
* Processes one event at a time in Run-to-Completion fashion.
* \param me is the pointer the state machine structure derived from ::QHsm.
*
* \note Must be called after QHsm_init().
*/
void QHsm_dispatch(QHsm *me);

/* protected methods... */

/** \brief The top-state.
*
* QHsm_top() is the ultimate root of state hierarchy in all HSMs derived
* from ::QHsm. This state handler always returns (QSTATE)0, which means
* that it "handles" all events.
*
* \sa Example of the QCalc_on() state handler for Q_INIT().
*/
QSTATE QHsm_top(QHsm *me);

/** \return the current active state of the \a me_ state machine  */
#define QHsm_getState_(me_) ((QHsmState const)(me_)->super_.state__)

#endif                                                            /* Q_NHSM */


/****************************************************************************/
#ifndef QF_TIMEEVT_CTR_SIZE
    /** \brief macro to override the default QTimeEvtCtr size.
    * Valid values 0, 1, 2, or 4; default 0
    */
    #define QF_TIMEEVT_CTR_SIZE 0
#endif
#if (QF_TIMEEVT_CTR_SIZE == 0)
#elif (QF_TIMEEVT_CTR_SIZE == 1)
    typedef uint8_t QTimeEvtCtr;
#elif (QF_TIMEEVT_CTR_SIZE == 2)
    /** \brief type of the Time Event counter, which determines the dynamic
    * range of the time delays measured in clock ticks.
    *
    * This typedef is configurable via the preprocessor switch
    * #QF_TIMEEVT_CTR_SIZE. The other possible values of this type are
    * as follows: \n
    * none when (QF_TIMEEVT_CTR_SIZE not defined or == 0), \n
    * uint8_t when (QF_TIMEEVT_CTR_SIZE == 1); \n
    * uint16_t when (QF_TIMEEVT_CTR_SIZE == 2); and \n
    * uint32_t when (QF_TIMEEVT_CTR_SIZE == 4).
    */
    typedef uint16_t QTimeEvtCtr;
#elif (QF_TIMEEVT_CTR_SIZE == 4)
    typedef uint32_t QTimeEvtCtr;
#else
    #error "QF_TIMER_SIZE defined incorrectly, expected 1, 2, or 4"
#endif

/** \brief a name for QActiveTag struct
*
* QActive is the base structure for derivation of active objects. Active
* objects in QF-nano are encapsulated tasks (each embedding a state machine
* and an event queue) that communicate with one another asynchronously by
* sending and receiving events. Within an active object, events are
* processed sequentially in a run-to-completion (RTC) fashion, while QF
* encapsulates all the details of thread-safe event exchange and queuing.
*
* \note ::QActive is not intended to be instantiated directly, but rather
* serves as the base structure for derivation of active objects in the
* application code.
*
* The following example illustrates how to derive an active object from
* QActive. Please note that the QActive member super_ is defined as the
* FIRST member of the derived struct.
* \include qfn_qactive.c
*
* \sa ::QActiveTag for the description of the data members \n \ref derivation
*/
typedef struct QActiveTag QActive;
struct QActiveTag {

#if (!defined(QF_FSM_ACTIVE) && !defined(Q_NHSM))
    QHsm super_;                /**< derives from the ::QHsm base structure */
#else
    QFsm super_;                /**< derives from the ::QFsm base structure */
#endif

    /** \brief offset to where next event will be inserted into the buffer
    */
    uint8_t head__;

    /** \brief offset of where next event will be extracted from the buffer
    */
    uint8_t tail__;

    /** \brief number of events currently present in the ring buffer
    */
    uint8_t nUsed__;

#if (QF_TIMEEVT_CTR_SIZE != 0)
    /** \brief Time Event tick counter for the active object
    */
    QTimeEvtCtr tickCtr__;
#endif
};

#if (Q_PARAM_SIZE != 0)
    /** \brief Posts an event \a e directly to the event queue of the acitve
    * object \a prio using the First-In-First-Out (FIFO) policy. This
    * function briefly locks and unlocks interrupts to protect the
    * queue integrity.
    *
    * Direct event posting is the only asynchronous communication method
    * available in QF-nano. The following example illustrates how the
    * Ped active object posts directly the PED_WAITING event to the PELICAN
    * crossing active object.
    * \include qfn_post.c
    *
    * \note The producer of the event (Ped in this case) must only "know"
    * the recipient's priority (Pelican), but the specific definition of
    * the Pelican structure is not required.
    *
    * \note Direct event posting should not be confused with direct event
    * dispatching. In contrast to asynchronous event posting through event
    * queues, direct event dispatching is synchronous. Direct event
    * dispatching occurs when you call QHsm_dispatch(), or QFsm_dispatch()
    * function.
    */
    void QF_post(uint8_t prio, QSignal sig, QParam par);

    /** \brief Posts an event \a e directly to the event queue of the acitve
    * object \a prio using the First-In-First-Out (FIFO) policy. This
    * function does NOT lock/unlock interrupts and is intended only
    * to be used inside critical sections (such as inside ISRs that cannot
    * nest).
    *
    * \sa QF_post()
    */
    void QF_postNoLock(uint8_t prio, QSignal sig, QParam par);
#else
    void QF_post(uint8_t prio, QSignal sig);
    void QF_postNoLock(uint8_t prio, QSignal sig);
#endif

/****************************************************************************/
/** \brief a name for QActive Control Block (CB) struct
*
* QActiveCB represents the constant information that the QF-nano needs
* to manage the active object. QActiveCB objects are grouped in the
* array QF_active[], which typically can be placed in ROM.
*
* The following example illustrates how to allocate and initialize the
* QActive control blocks in the array QF_active[].
* \include qfn_main.c
*/
typedef struct QActiveCBTag QActiveCB;

/** \brief QActive Control Block
* \sa ::QActiveCB
*/
struct QActiveCBTag {
    QActive *act__;      /**< \brief pointer to the active object structure */
    QEvent *queue__;          /**< \brief pointer to the event queue buffer */
    uint8_t end__;                /**< \brief the length of the ring buffer */
};

#if (QF_TIMEEVT_CTR_SIZE != 0)

#if (QF_TIMEEVT_CTR_SIZE > 1)

/** \brief Arm a one-shot time event for direct event posting.
*
* Arms a time event \a me to fire in \a tout clock ticks
* (one-shot time event). The timeout signal Q_TIMEOUT_SIG gets directly
* posted (using the FIFO policy) into the event queue of the active object
* \a me.
*
* After posting, the time event gets automatically disarmed and can be reused.
*
* A one-shot time event can be disarmed at any time by calling the
* QActive_disarm() function. Also, a one-shot time event can be re-armed
* to fire in a different number of clock ticks by calling QActive_arm() again.
*
* The following example shows how to arm a one-shot time event from a state
* machine of an active object:
* \include qfn_arm.c
*/
void QActive_arm(QActive *me, QTimeEvtCtr tout);

/** \brief Disarm a time event.
*
* The time event \a me gets disarmed and can be reused.
*/
void QActive_disarm(QActive *me);

#else

/** \brief Arm a one-shot time event for direct event posting.
*
* Arms a time event \param me_ to fire in \param tout_ clock ticks
* (one-shot time event). The timeout signal Q_TIMEOUT_SIG gets directly
* posted (using the FIFO policy) into the event queue of the active object
* \param me_.
*
* After posting, the time event gets automatically disarmed and can be reused.
*
* A one-shot time event can be disarmed at any time by calling the
* QActive_disarm() function. Also, a one-shot time event can be re-armed
* to fire in a different number of clock ticks by calling QActive_arm() again.
*
* The following example shows how to arm a one-shot time event from a state
* machine of an active object:
* \include qfn_arm.c
*/
#define QActive_arm(me_, tout_) ((me_)->tickCtr__ = (QTimeEvtCtr)(tout_))

/** \brief Disarm a time event.
*
* The time event \param me_ gets disarmed and can be reused.
*/
#define QActive_disarm(me_)     ((me_)->tickCtr__ = (QTimeEvtCtr)0)

#endif                                           /* QF_TIMEEVT_CTR_SIZE > 1 */

/** \brief Processes all armed time events at every clock tick.
*
* This function must be called periodically from a time-tick ISR or from
* the highest-priority task so that QF can manage the timeout events.
*
* \note The QF_tick() function is not reentrant meaning that it must run to
* completion before it is called again. Also, QF_tick() assumes that it
* never will get preempted by a task, which is always the case when it is
* called from an ISR or the highest-priority task.
*
* The following example illustrates the call to QF_tick():
* \include qfn_tick.c
*/
void QF_tick(void);

#endif

/* protected methods ...*/

/** \brief QF initialization.
*
* This function initializes QF and must be called exactly once before any
* other QF function. In QF-nano this function is defined in the BSP.
*/
void QF_init(void);

/** \brief Starts the interrupts and initializes other critical resources
* that might interact with the QF application.
*
* QF_start() is called from QF_run(), right before starting the non-preemptive
* multitasking in the background loop.
*
* \note This function is strongly platform-dependent and is not implemented
* in the QF, but either in the QF port or in the Board Support Package (BSP)
* for the given application.
*
* \sa QF initialization example for ::QActiveCB.
*/
void QF_start(void);

/** \brief Transfers control to QF to run the application.
*
* QF_run() implemetns the simple non-preemptive scheduler. QF_run() must be
* called from your startup code after you initialize the QF and define at
* least one active object control block in QF_active[].
*
* \note When the Quantum Kernel (QK) is used as the underlying real-time
* kernel for the QF, all platfrom dependencies are handled in the QK, so
* no porting of QF is necessary. In other words, you only need to recompile
* the QF platform-independent code with the compiler for your platform, but
* you don't need to provide any platform-specific implementation (so, no
* qf_port.c file is necessary). Moreover, QK implements the function QF_run()
* in a platform-independent way, in the modile qk.c.
*/
void QF_run(void);

/** \brief QF idle callback (customized in BSPs for QF)
*
* QF_onIdle() is called is some QF ports when the QF detects that no events
* are available for active objects. This callback gives the application an
* opportunity to enter a power-saving CPU mode, or perform some other
* idle processing (such as Q-Spy output).
*/
void QF_onIdle(void);

/** \brief Exits the QF application and returns control to the OS/Kernel.
*
* This function exits the framework. After calling this function, QF is no
* longer in control of the application. The typical use of this method is
* for exiting the QF application to return back to the operating system
* or for handling fatal errors that require resetting the system.
*
* This function is strongly platform-dependent and is not implemented in
* QF-nano, but either in the QF port or in the Board Support Package (BSP)
* for the given application. Some QF ports might not require implementing
* QF_exit() at all, because many embedded application don't have anything
* to exit to.
*/
void QF_exit(void);


extern QActiveCB const Q_ROM QF_active[]; /**< active object control blocks */
extern uint8_t   const Q_ROM QF_activeNum;/**< the number of control blocks */

/****************************************************************************/
/** \page qpn_rev QP-nano Revision History

\section qpn_1_1_00 Version 1.1.00 (Commercial Release)
Release date: May 15, 2006\n

-# First commercial release.

*/

#endif                                                             /* qpn_h */
